/* 
 * Realign accesses to misaligned ethernet frame.
 *
 * Copyright (c) 2005 TiVo Inc.
 * 
 * This software may be used and distributed according to the terms
 * of the GNU General Public License, incorporated herein by reference.
 *
 * Also see:
 *   include/asm-mips/checksum.h
 *   include/linux/icmp.h
 *   include/linux/ip.h
 *   include/linux/tcp.h
 *   include/net/sock.h
 *   include/net/realign.h
 *   drivers/net/Config.in
 *   drivers/net/bcmenet.c
 *   net/ipv4/ipconfig.c
 *   net/ipv4/tcp_input.c
 *   net/sunrpc/xdr.c
 *   net/sunrpc/xprt.c
 */
  
#ifndef _NET_REALIGN_H
#define _NET_REALIGN_H

#include <linux/config.h>

#ifdef CONFIG_TIVO_REALIGN_ETHERNET

#include <asm/unaligned.h>

/* 
 * This is used in tcp_input.c and friends, and fixes various direct 32-bit
 * accesses to the skb, ie: 
 *   * (u32 *) skb_data   =>  fetch_unalign((u32 *) skb_data)
 */ 
#define fetch_realign(p) get_unaligned(p)

/* 
 * This is applied to various struct definitions, it causes gcc to assume that
 * accesses to a 'packed' structure must be unaligned.
 */ 
#define header_realign __attribute__((packed))

#else

/* here, ethernet is aligned by hardware, do no harm */ 
#define fetch_realign(p) *(p)
#define header_realign

#endif

#endif
